import hashlib
from uuid import UUID

from pyramid.settings import aslist

from kinto.core import utils as core_utils


def is_monitoring_collection(registry, bucket_id, collection_id):
    resources_uri = aslist(registry.settings.get('changes.resources', ''))

    for resource_uri in resources_uri:
        resource_name, matchdict = core_utils.view_lookup_registry(registry, resource_uri)
        if resource_name == 'bucket' and bucket_id == matchdict['id']:
            return True

        is_matching_collection = (
            resource_name == 'collection' and
            bucket_id == matchdict['bucket_id'] and
            collection_id == matchdict['id']
        )
        if is_matching_collection:
            return True


def monitored_collections(registry):
    storage = registry.storage
    resources_uri = aslist(registry.settings.get('changes.resources', ''))
    collections = []

    for resource_uri in resources_uri:
        resource_name, matchdict = core_utils.view_lookup_registry(registry, resource_uri)
        if resource_name == 'bucket':
            # Every collections in this bucket.
            result = storage.list_all(resource_name='collection', parent_id=resource_uri)
            collections.extend([(matchdict['id'], obj['id']) for obj in result])

        elif resource_name == 'collection':
            collections.append((matchdict['bucket_id'], matchdict['id']))

    return collections


def changes_object(request, bucket_id, collection_id, timestamp):
    """Generate an object for /buckets/monitor/collections/changes."""
    http_host = request.registry.settings.get('http_host') or ''
    collection_uri = core_utils.instance_uri(
        request, 'collection', bucket_id=bucket_id, id=collection_id)
    uniqueid = http_host + collection_uri
    identifier = hashlib.md5(uniqueid.encode('utf-8')).hexdigest()
    entry_id = str(UUID(identifier))

    return dict(id=entry_id,
                last_modified=timestamp,
                bucket=bucket_id,
                collection=collection_id,
                host=http_host)



import os
import configparser

from kinto import main as kinto_main
from kinto.core.testing import get_user_headers, BaseWebTest as CoreWebTest


here = os.path.abspath(os.path.dirname(__file__))


class BaseWebTest(CoreWebTest):
    api_prefix = "v1"
    entry_point = kinto_main
    config = 'config.ini'

    def __init__(self, *args, **kwargs):
        super(BaseWebTest, self).__init__(*args, **kwargs)
        self.headers.update(get_user_headers('mat'))
        self.headers.update({'Origin': 'http://localhost:9999'})

    @classmethod
    def get_app_settings(cls, extras=None):
        ini_path = os.path.join(here, cls.config)
        config = configparser.ConfigParser()
        config.read(ini_path)
        settings = dict(config.items('app:main'))
        settings.update(extras or {})
        return settings

    def setUp(self):
        super(BaseWebTest, self).setUp()
        self.create_collection('blocklists', 'certificates')

    def create_collection(self, bucket_id, collection_id):
        bucket_uri = '/buckets/%s' % bucket_id
        self.app.put_json(bucket_uri,
                          {},
                          headers=self.headers)
        collection_uri = bucket_uri + '/collections/%s' % collection_id
        self.app.put_json(collection_uri,
                          {},
                          headers=self.headers)



import unittest
from unittest import mock

from pyramid.request import Request

from kinto_changes.utils import changes_object


class ChangesRecordTest(unittest.TestCase):
    def test_single_hardcoded(self):
        request = Request.blank(path='/')
        request.route_path = mock.Mock()
        request.route_path.return_value = '/buckets/a/collections/b'
        request.registry = mock.Mock()
        request.registry.settings = {}
        timestamp = 1525457597166
        entry = changes_object(request, 'a', 'b', timestamp)

        self.assertEqual(entry, {
            "bucket": "a",
            "collection": "b",
            "host": "",
            "id": "9527d115-6191-fa49-a530-8fbfc4997755",
            "last_modified": timestamp
        })

    def test_another_hardcoded(self):
        request = Request.blank(path='/')
        request.route_path = mock.Mock()
        request.route_path.return_value = '/buckets/a/collections/b'
        request.registry = mock.Mock()
        request.registry.settings = {'http_host': 'https://localhost:443'}
        timestamp = 1525457597166
        entry = changes_object(request, 'a', 'b', timestamp)

        self.assertEqual(entry, {
            "bucket": "a",
            "collection": "b",
            "host": "https://localhost:443",
            "id": "fa48a96d-1600-f561-8645-3395acb08a5a",
            "last_modified": timestamp
        })



import hashlib
import unittest
from unittest import mock

from uuid import UUID

from . import BaseWebTest
from kinto_changes import includeme

SAMPLE_RECORD = {'data': {'dev-edition': True}}


def monitor_changes_events(events):
    return [e for e in events
            if e.payload.get('bucket_id') == 'monitor' and
            e.payload.get('collection_id') == 'changes']


class RedirectEventsTest(BaseWebTest):
    changes_uri = '/buckets/monitor/collections/changes/records'
    records_uri = '/buckets/blocklists/collections/certificates/records'

    @classmethod
    def make_app(cls, *args, **kwargs):
        with mock.patch('tests.listener.load_from_config') as listener_loader:
            cls.listener = listener_loader.return_value
            return super(RedirectEventsTest, cls).make_app(*args, **kwargs)

    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super(RedirectEventsTest, cls).get_app_settings(extras)
        settings['event_listeners'] = 'tests.listener'
        settings['kinto.changes.resources'] = ' '.join([
            '/buckets/blocklists',
            '/buckets/fennec/collections/fonts'
        ])
        return settings

    def setUp(self):
        super(RedirectEventsTest, self).setUp()
        self.registry = self.app.app.registry
        self.listener.call_args_list = []

    def assert_event_is_for_collection(self, event, bucket_id, collection_id):
        uniqueid = 'www.kinto-storage.org/buckets/{}/collections/{}'.format(
            bucket_id, collection_id)
        event_id = hashlib.md5(uniqueid.encode('utf-8')).hexdigest()
        event_id = str(UUID(event_id))

        self.assertEqual(len(event.impacted_records), 1)
        impacted_record = event.impacted_records[0]
        new = impacted_record['new']
        self.assertEqual(new['id'], event_id)
        self.assertEqual(new['bucket'], bucket_id)
        self.assertEqual(new['collection'], collection_id)


class BasicEventsTest(RedirectEventsTest, unittest.TestCase):
    def test_generate_event(self):
        self.app.post_json(self.records_uri, SAMPLE_RECORD,
                           headers=self.headers)
        events = [call[0][0] for call in self.listener.call_args_list]
        changes_events = monitor_changes_events(events)
        self.assertEqual(len(changes_events), 1)
        event = changes_events[0]

        self.assert_event_is_for_collection(event, 'blocklists', 'certificates')

    def test_ignore_event_for_unmonitored(self):
        self.create_collection('foo', 'certificates')
        self.app.post_json('/buckets/foo/collections/certificates/records',
                           SAMPLE_RECORD, headers=self.headers)
        events = [call[0][0] for call in self.listener.call_args_list]
        changes_events = monitor_changes_events(events)
        self.assertEqual(len(changes_events), 0)

    def test_event_for_collections(self):
        self.create_collection('fennec', 'fonts')
        self.app.post_json('/buckets/fennec/collections/fonts/records', SAMPLE_RECORD,
                           headers=self.headers)
        events = [call[0][0] for call in self.listener.call_args_list]
        changes_events = monitor_changes_events(events)
        self.assertEqual(len(changes_events), 1)
        event = changes_events[0]

        self.assert_event_is_for_collection(event, 'fennec', 'fonts')


class InitializationTest(RedirectEventsTest, unittest.TestCase):
    def test_scan_timestamps_on_startup(self):
        self.create_collection('blocklists', 'pinning')
        res = self.app.get('/buckets/blocklists/collections/pinning/records', headers=self.headers)
        last_modified = int(res.headers['ETag'][1:-1])

        old_storage = self.storage
        old_permission = self.permission

        def restore_storage(config):
            """Bring old app's storage/permission (memory-based) over to a new app.

            This lets us test initialization in the case of existing collections."""
            config.registry.storage = old_storage
            config.registry.permission = old_permission

        with mock.patch('tests.test_events', restore_storage):
            # Careful -- this covers cls.listener and can lead to some
            # surprising behavior if you add another test to this class
            app = self.make_app({
                'kinto.includes': 'tests.test_events kinto_changes'
            })

        app.post_json('/buckets/blocklists/collections/pinning/records', SAMPLE_RECORD,
                      headers=self.headers)
        events = [call[0][0] for call in self.listener.call_args_list]
        monitor_changes_events = [e for e in events
                                  if e.payload.get('bucket_id') == 'monitor' and
                                  e.payload.get('collection_id') == 'changes']
        self.assertEqual(len(monitor_changes_events), 1)
        event = monitor_changes_events[0]

        self.assert_event_is_for_collection(event, 'blocklists', 'pinning')
        self.assertEqual(event.impacted_records[0]['old']['last_modified'], last_modified)


class StatsdTest(BaseWebTest, unittest.TestCase):
    def test_statsd_wraps_subscriber(self):
        config = mock.MagicMock()
        config.settings = {}
        statsd = config.registry.statsd
        statsd_decorate = statsd.timer.return_value

        includeme(config)
        self.assertEqual(statsd.timer.call_args_list[0][0][0], 'plugins.kinto_changes')
        # Check that we're monitoring some calls.
        self.assertTrue(statsd_decorate.call_count >= 2)


class DontMigrateTest(BaseWebTest, unittest.TestCase):
    def test_migrate_does_nothing(self):
        config = mock.MagicMock()
        config.settings = {}
        config.registry.command = 'migrate'

        includeme(config)
        self.assertNotIn('changes', config.capabilities)



"""Listener that mirrors events from any collection into the changes collection."""

from kinto.core import utils as core_utils
from kinto.core.events import notify_resource_event, ACTIONS
from .utils import monitored_collections, changes_object, is_monitoring_collection


class Listener(object):
    def __init__(self, config):
        self.registry = config.registry
        self.collection_timestamps = {}

    def track_timestamps(self, event):
        for (bucket_id, collection_id) in monitored_collections(self.registry):
            timestamp = self.get_collection_timestamp(bucket_id, collection_id)
            self.collection_timestamps[(bucket_id, collection_id)] = timestamp

    def get_collection_timestamp(self, bucket_id, collection_id):
        collection_uri = core_utils.instance_uri_registry(self.registry,
                                                          'collection',
                                                          bucket_id=bucket_id,
                                                          id=collection_id)
        return self.registry.storage.resource_timestamp(parent_id=collection_uri,
                                                        resource_name='record')

    def on_record_changed(self, event):
        bucket_id = event.payload['bucket_id']
        collection_id = event.payload['collection_id']
        timestamp = self.get_collection_timestamp(bucket_id, collection_id)

        if not is_monitoring_collection(self.registry, bucket_id, collection_id):
            return

        for change in event.impacted_records:
            # This might be the first we've seen of this collection.
            old_timestamp = self.collection_timestamps.get((bucket_id, collection_id), None)

            # Synthesize event for /buckets/monitor/collections/changes record.
            new = changes_object(event.request, bucket_id, collection_id, timestamp)
            old = None
            action = ACTIONS.CREATE
            if old_timestamp:
                action = ACTIONS.UPDATE
                old = changes_object(event.request, bucket_id, collection_id, old_timestamp)

            resource_data = {'bucket_id': 'monitor', 'collection_id': 'changes', 'id': new['id']}

            notify_resource_event(
                event.request, '/buckets/monitor/collections/changes',
                timestamp, new, action, old=old,
                resource_name='record', resource_data=resource_data
            )

            self.collection_timestamps[(bucket_id, collection_id)] = timestamp



import pkg_resources
import pyramid.events
from pyramid.settings import aslist
from kinto.core import events
from .listener import Listener

#: Module version, as defined in PEP-0396.
__version__ = pkg_resources.get_distribution(__package__).version

MONITOR_BUCKET = 'monitor'
MONITOR_BUCKET_PATH = '/buckets/{}'.format(MONITOR_BUCKET)
CHANGES_COLLECTION = 'changes'
CHANGES_COLLECTION_PATH = '{}/collections/{}'.format(
    MONITOR_BUCKET_PATH, CHANGES_COLLECTION)
CHANGES_RECORDS_PATH = '{}/records'.format(CHANGES_COLLECTION_PATH)
CHANGESET_PATH = '/buckets/{bid}/collections/{cid}/changeset'


def identity(f):
    """Used as a fallback decorator if no statsd configured."""
    return f


def includeme(config):
    settings = config.get_settings()
    collections = settings.get('changes.resources', [])

    # Don't do anything on a migration command.
    # This is helpful because we try to get a snapshot of the database
    # when the application is created. But if we're doing a migrate,
    # there may not be a database.
    if getattr(config.registry, 'command', None) == 'migrate':
        return

    config.add_api_capability(
        "changes",
        version=__version__,
        description="Track modifications of records in Kinto and store"
                    " the collection timestamps into a specific bucket"
                    " and collection.",
        url="http://kinto.readthedocs.io/en/latest/tutorials/"
        "synchronisation.html#polling-for-remote-changes",
        collections=aslist(collections))

    config.scan('kinto_changes.views')

    listener = Listener(config)

    decorate = identity
    if config.registry.statsd:
        key = 'plugins.kinto_changes'
        decorate = config.registry.statsd.timer(key)

    config.add_subscriber(decorate(listener.track_timestamps),
                          pyramid.events.ApplicationCreated)

    config.add_subscriber(decorate(listener.on_record_changed),
                          events.ResourceChanged,
                          for_resources=('record'))
